import numpy as np
import pytest

import tiledb
from tiledb import TileDBError
from tiledb.tests.common import DiskTestCase


class SubarrayTest(DiskTestCase):
    def test_add_range(self):
        dim1 = tiledb.Dim("row", domain=(1, 10))
        dim2 = tiledb.Dim("col", domain=(1, 10))
        dom = tiledb.Domain(dim1, dim2)
        att = tiledb.Attr("val", dtype=np.uint64)
        schema = tiledb.ArraySchema(domain=dom, attrs=(att,))
        uri = self.path("dense_array")
        tiledb.Array.create(uri, schema)

        with tiledb.open(uri, "w") as array:
            array[1:5, 1:5] = np.reshape(np.arange(1, 17, dtype=np.float64), (4, 4))

        with tiledb.open(uri, "r") as array:
            subarray1 = tiledb.Subarray(array)

            # Check number of ranges: each dimension should have the default range.
            assert subarray1.num_dim_ranges(0) == 1
            assert subarray1.num_dim_ranges(1) == 1
            assert subarray1.shape() == (10, 10)

            # Add range to first dim and check still only 1 range (replace default).
            subarray1.add_dim_range(0, (1, 2))
            assert subarray1.num_dim_ranges(0) == 1
            assert subarray1.shape() == (2, 10)

            # Add additional range to first dim and check 2 ranges.
            subarray1.add_dim_range(0, (4, 4))
            assert subarray1.num_dim_ranges(0) == 2
            assert subarray1.shape() == (3, 10)

    def test_add_ranges_basic(self):
        uri = self.path("test_pyquery_basic")
        with tiledb.from_numpy(uri, np.random.rand(4)):
            pass

        with tiledb.open(uri) as array:
            subarray = tiledb.Subarray(array)

            subarray.add_ranges([[(0, 3)]])

            with self.assertRaises(TileDBError):
                subarray.add_ranges([[(0, 3.0)]])

            subarray.add_ranges([[(0, np.int32(3))]])

            with self.assertRaises(TileDBError):
                subarray.add_ranges([[(3, "a")]])

            with self.assertRaisesRegex(
                TileDBError,
                "Failed to cast dim range '\\(1.2344, 5.6789\\)' to dim type UINT64.*$",
            ):
                subarray.add_ranges([[(1.2344, 5.6789)]])

            with self.assertRaisesRegex(
                TileDBError,
                "Failed to cast dim range '\\('aa', 'bbbb'\\)' to dim type UINT64.*$",
            ):
                subarray.add_ranges([[("aa", "bbbb")]])

    @pytest.mark.skipif(
        tiledb.libtiledb.version()[0] == 2 and tiledb.libtiledb.version()[1] < 15,
        reason="dimension labels requires libtiledb version 2.15 or greater",
    )
    def test_add_label_ranges_1d(self):
        # Create array schema with dimension labels
        dim = tiledb.Dim("d1", domain=(1, 10), dtype=np.uint32)
        dom = tiledb.Domain(dim)
        att = tiledb.Attr("a1", dtype=np.int64)
        dim_labels = {0: {"l1": dim.create_label_schema("increasing", np.int64)}}
        schema = tiledb.ArraySchema(domain=dom, attrs=(att,), dim_labels=dim_labels)

        # Create array
        uri = self.path("dense_array_with_label")
        tiledb.Array.create(uri, schema)

        # Add dimension label ranges
        with tiledb.open(uri, "r") as array:
            subarray1 = tiledb.Subarray(array)
            assert subarray1.num_dim_ranges(0) == 1

            subarray1.add_label_range("l1", (-1, 1))
            assert subarray1.num_dim_ranges(0) == 0
            assert subarray1.num_label_ranges("l1") == 1

    def test_copy_ranges(self):
        # Create array schema with dimension labels
        d1 = tiledb.Dim("d1", domain=(1, 10), dtype=np.uint32)
        d2 = tiledb.Dim("d2", domain=(1, 10), dtype=np.uint32)
        d3 = tiledb.Dim("d3", domain=(1, 10), dtype=np.uint32)
        d4 = tiledb.Dim("d4", domain=(1, 10), dtype=np.uint32)
        dom = tiledb.Domain(d1, d2, d3, d4)
        att = tiledb.Attr("a1", dtype=np.int64)
        schema = tiledb.ArraySchema(domain=dom, attrs=(att,))

        # Create array
        uri = self.path("array4d")
        tiledb.Array.create(uri, schema)

        # Add ranges (1, 1) and (3, 3) to each dimension of the subarray.
        with tiledb.open(uri, "r") as array:
            subarray1 = tiledb.Subarray(array)
            subarray1.add_ranges(
                (((1, 1), (3, 3)), ((1, 1), (3, 3)), ((1, 1), (3, 3)), ((1, 1), (3, 3)))
            )
            assert subarray1.num_dim_ranges(0) == 2
            assert subarray1.num_dim_ranges(1) == 2
            assert subarray1.num_dim_ranges(2) == 2
            assert subarray1.num_dim_ranges(3) == 2

            # Should copy ranges from d1 and d3.
            # All other dimensions should only have default range.
            subarray2 = tiledb.Subarray(array)
            subarray2.copy_ranges(subarray1, [0, 2])
            assert subarray2.num_dim_ranges(0) == 2
            assert subarray2.num_dim_ranges(1) == 1
            assert subarray2.num_dim_ranges(2) == 2
            assert subarray2.num_dim_ranges(3) == 1



import json
import xml

import numpy as np
import pytest

import tiledb
from tiledb.tests.common import DiskTestCase


class ConsolidationPlanTest(DiskTestCase):
    def test_consolidation_plan(self):
        path = self.path("test_consolidation_plan")

        array = np.random.rand(4)
        tiledb.from_numpy(path, array)

        with tiledb.open(path, "r") as A:
            cons_plan = tiledb.ConsolidationPlan(tiledb.default_ctx(), A, 2)
            assert cons_plan.num_nodes == 1
            assert cons_plan.num_nodes == len(cons_plan)
            assert cons_plan.num_fragments(0) == 1
            # check that it has a nodes key
            assert "nodes" in json.loads(cons_plan.dump())
            # check that it has a list of nodes
            assert isinstance(json.loads(cons_plan.dump())["nodes"], list)
            # check that each node has a uri key
            for node in json.loads(cons_plan.dump())["nodes"]:
                assert "uri" in node["uris"][0]
            # test __repr__
            try:
                assert (
                    xml.etree.ElementTree.fromstring(cons_plan._repr_html_())
                    is not None
                )
            except:
                pytest.fail(
                    f"Could not parse cons_plan._repr_html_(). Saw {cons_plan._repr_html_()}"
                )
            # test __getitem__
            assert cons_plan[0] == {
                "num_fragments": 1,
                "fragment_uris": [cons_plan.fragment_uri(0, 0)],
            }

        # write a second fragment to the array and check the new consolidation plan
        with tiledb.open(path, "w") as A:
            A[:] = np.random.rand(4)

        with tiledb.open(path, "r") as A:
            cons_plan = tiledb.ConsolidationPlan(tiledb.default_ctx(), A, 4)
            assert cons_plan.num_nodes == 1
            assert cons_plan.num_fragments(0) == 2



import os

import pytest

import tiledb
import tiledb.libtiledb as lt


@pytest.fixture
def text_fname(tmp_path):
    path = os.path.join(tmp_path, "text_fname")
    vfs = tiledb.VFS()
    vfs.touch(path)
    with vfs.open(path, "wb") as fio:
        fio.write(b"Simple text file.\n")
        fio.write(b"With two lines.")
    return path


def test_lt_schema_create(text_fname):
    ctx = lt.Context()
    schema = lt.Filestore._schema_create(ctx, text_fname)
    assert type(schema) == lt.ArraySchema


def test_libtiledb_schema_create_buffer(tmp_path, text_fname):
    ctx = lt.Context()
    path = os.path.join(tmp_path, "test_libtiledb_schema_create_buffer")
    schema = tiledb.ArraySchema.from_file(text_fname)
    tiledb.Array.create(path, schema)

    data = b"buffer"
    lt.Filestore._buffer_import(ctx, path, data, lt.MIMEType.AUTODETECT)
    assert bytes(data) == lt.Filestore._buffer_export(ctx, path, 0, len(data))
    assert lt.Filestore._size(ctx, path) == len(data)

    output_file = os.path.join(tmp_path, "output_file")
    vfs = tiledb.VFS()
    vfs.touch(output_file)
    lt.Filestore._uri_export(ctx, path, output_file)
    with vfs.open(output_file, "rb") as fio:
        assert fio.read() == data


def test_libtiledb_schema_create_uri(tmp_path, text_fname):
    ctx = lt.Context()
    path = os.path.join(tmp_path, "test_libtiledb_schema_create_uri")
    schema = tiledb.ArraySchema.from_file(text_fname)
    tiledb.Array.create(path, schema)

    lt.Filestore._uri_import(ctx, path, text_fname, lt.MIMEType.AUTODETECT)
    with open(text_fname, "rb") as text:
        data = text.read()
        assert data == lt.Filestore._buffer_export(ctx, path, 0, len(data))
        assert lt.Filestore._size(ctx, path) == len(data)


def test_mime_type():
    to_str = {
        lt.MIMEType.AUTODETECT: "AUTODETECT",
        lt.MIMEType.TIFF: "image/tiff",
        lt.MIMEType.PDF: "application/pdf",
    }

    for k in to_str:
        assert lt.Filestore._mime_type_to_str(k) == to_str[k]

    from_str = {
        "AUTODETECT": lt.MIMEType.AUTODETECT,
        "image/tiff": lt.MIMEType.TIFF,
        "application/pdf": lt.MIMEType.PDF,
    }

    for k in from_str:
        assert lt.Filestore._mime_type_from_str(k) == from_str[k]



# reading_dense_layouts.py
#
# LICENSE
#
# The MIT License
#
# Copyright (c) 2020 TileDB, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# DESCRIPTION
#
# Please see the TileDB documentation for more information:
#   https://docs.tiledb.com/main/how-to/arrays/reading-arrays/basic-reading
#
# When run, this program will create a simple 2D dense array, write some data
# to it, and read a slice of the data back in the layout of the user's choice
# (passed as an argument to the program: "row", "col", or "global").
#

import sys

import numpy as np

import tiledb

# Name of the array to create.
array_name = "reading_dense_layouts"


def create_array():
    # The array will be 4x4 with dimensions "rows" and "cols", with domain [1,4].
    dom = tiledb.Domain(
        tiledb.Dim(name="rows", domain=(1, 4), tile=2, dtype=np.int32),
        tiledb.Dim(name="cols", domain=(1, 4), tile=2, dtype=np.int32),
    )

    # The array will be dense with a single attribute "a" so each (i,j) cell can store an integer.
    schema = tiledb.ArraySchema(
        domain=dom, sparse=False, attrs=[tiledb.Attr(name="a", dtype=np.int32)]
    )

    # Create the (empty) array on disk.
    tiledb.DenseArray.create(array_name, schema)


def write_array():
    # Open the array and write to it.
    with tiledb.DenseArray(array_name, mode="w") as A:
        # NOTE: global writes are not currently supported in the Python API.
        # The following code will produce the same array as the corresponding
        # C++ example in the docs (which wrote in global order)
        data = np.array(([1, 2, 5, 6], [3, 4, 7, 8], [9, 10, 13, 14], [11, 12, 15, 16]))
        A[:] = data


def read_array(order):
    # Open the array and read from it.
    with tiledb.DenseArray(array_name, mode="r") as A:
        # Get non-empty domain
        print("Non-empty domain: {}".format(A.nonempty_domain()))

        # Slice only rows 1, 2 and cols 2, 3, 4.
        # NOTE: The `query` syntax is required to get the coordinates for
        # dense arrays and specify an order other than the default row-major
        data = A.query(attrs=["a"], order=order, coords=True)[1:3, 2:5]
        a_vals = data["a"]
        coords = np.asarray(list(zip(data["rows"], data["cols"])))

        if order != "G" and a_vals.flags["F_CONTIGUOUS"]:
            print("NOTE: The following result array has col-major layout internally")

        if order != "G":
            for i in range(coords.shape[0]):
                for j in range(coords.shape[1]):
                    print(
                        "Cell {} has data {}".format(
                            str(coords[i, j]), str(a_vals[i, j])
                        )
                    )
        else:
            # When reading in global order, TileDB always returns a vector (1D array)
            for i in range(coords.shape[0]):
                print("Cell {} has data {}".format(str(coords[i]), str(a_vals[i])))


# Check if the array already exists.
if tiledb.object_type(array_name) != "array":
    create_array()
    write_array()

layout = ""
if len(sys.argv) > 1:
    layout = sys.argv[1]

order = "C"
if layout == "col":
    order = "F"
elif layout == "global":
    order = "G"
else:
    order = "C"

read_array(order)



from __future__ import annotations

import io
from typing import Any, Optional, Sequence

import numpy as np
from numpy.typing import NDArray

import tiledb.libtiledb as lt

from .ctx import Ctx, CtxMixin
from .datatypes import DataType


class Enumeration(CtxMixin, lt.Enumeration):
    """
    Represents a TileDB Enumeration.
    """

    def __init__(
        self,
        name: str,
        ordered: bool,
        values: Optional[Sequence[Any]] = None,
        dtype: Optional[np.dtype] = None,
        ctx: Optional[Ctx] = None,
    ):
        """Class representing the TileDB Enumeration.

        :param name: The name of the to-be created Enumeration
        :type name: str
        :param ordered: Whether or not to consider this enumeration ordered
        :type ordered: bool
        :param values: A Numpy array of values for this enumeration
        :type values: np.array
        :param dtype: The Numpy data type for this enumeration
        :type dtype: np.dtype
        :param ctx: A TileDB context
        :type ctx: tiledb.Ctx
        """
        if values is None or len(values) == 0:
            if dtype is None:
                raise ValueError("dtype must be provided for empty enumeration")
            super().__init__(ctx, name, np.dtype(dtype), ordered)

        values = np.array(values)
        if np.dtype(values.dtype).kind in "US":
            dtype = (
                lt.DataType.STRING_UTF8
                if values.dtype.kind == "U"
                else lt.DataType.STRING_ASCII
            )
            super().__init__(ctx, name, values, ordered, dtype)
        else:
            super().__init__(ctx, name, ordered, values, np.array([]))

    @property
    def name(self) -> str:
        """The enumeration label string.

        :rtype: str
        """
        return super().name

    @property
    def dtype(self) -> np.dtype:
        """Numpy dtype representation of the enumeration type.

        :rtype: numpy.dtype
        """
        return DataType.from_tiledb(super().type).np_dtype

    @property
    def cell_val_num(self) -> int:
        """The enumeration's cell value number.

        :rtype: int
        """
        return super().cell_val_num

    @property
    def ordered(self) -> bool:
        """True if the enumeration is ordered.

        :rtype: bool
        """
        return super().ordered

    def values(self) -> NDArray:
        """The values of the enumeration.

        :rtype: NDArray
        """
        if self.dtype.kind == "U":
            return np.array(super().str_values(), dtype=np.str_)
        elif self.dtype.kind == "S":
            return np.array(super().str_values(), dtype=np.bytes_)
        else:
            return np.array(super().values(), dtype=self.dtype)

    def extend(self, values: Sequence[Any]) -> Enumeration:
        """Add additional values to the enumeration.

        :param values: The values to add to the enumeration
        :rtype: Enumeration
        """
        values = np.array(values)
        if self.dtype.kind in "US" and values.dtype.kind not in "US":
            raise lt.TileDBError("Passed in enumeration must be string type")

        if np.issubdtype(self.dtype, np.integer) and not np.issubdtype(
            values.dtype, np.integer
        ):
            raise lt.TileDBError("Passed in enumeration must be integer type")

        return Enumeration.from_pybind11(self._ctx, super().extend(values))

    def __eq__(self, other):
        if not isinstance(other, Enumeration):
            return False

        return all(
            [
                self.name == other.name,
                self.dtype == other.dtype,
                self.cell_val_num == other.cell_val_num,
                self.ordered == other.ordered,
                np.array_equal(self.values(), other.values()),
            ]
        )

    def __repr__(self):
        # use safe repr if pybind11 constructor failed
        if self._ctx is None:
            return object.__repr__(self)

        return f"Enumeration(name='{self.name}', dtype={self.dtype}, dtype_name='{self.dtype.name}', cell_val_num={self.cell_val_num}, ordered={self.ordered}, values={list(self.values())})"

    def _repr_html_(self):
        output = io.StringIO()

        output.write("<table>")
        output.write("<tr>")
        output.write("<th>Name</th>")
        output.write("<th>Data Type</th>")
        output.write("<th>Ordered</th>")
        output.write("</tr>")
        output.write(f"{self._repr_html_row_only_()}")
        output.write("</table>")

        return output.getvalue()

    def _repr_html_row_only_(self):
        output = io.StringIO()

        output.write("<tr>")
        output.write(f"<td>{self.name}</td>")
        output.write(f"<td>{self.dtype}</td>")
        output.write(f"<td>{self.cell_val_num}</td>")
        output.write(f"<td>{self.ordered}</td>")
        output.write("</tr>")

        return output.getvalue()



import itertools
import re
import textwrap
import warnings

import numpy as np

import tiledb

from .common import (
    DiskTestCase,
    fx_sparse_cell_order,  # noqa: F401
)


class ReprTest(DiskTestCase):
    def test_attr_repr(self):
        attr = tiledb.Attr(name="itsanattr", dtype=np.float64)
        self.assertTrue(
            re.match(
                r"Attr\(name=[u]?'itsanattr', dtype='float64', var=False, nullable=False, enum_label=None\)",
                repr(attr),
            )
        )

        g = dict()
        exec("from tiledb import Attr; from numpy import float64", g)
        self.assertEqual(eval(repr(attr), g), attr)

    def test_dim_repr(self):
        dtype_set = [bytes, np.bytes_]
        opts = {
            None: None,
            "var": True,
            "domain": (None, None),
            "filters": [tiledb.GzipFilter()],
        }

        dim_test_imports = textwrap.dedent(
            """
            from tiledb import Dim, FilterList, GzipFilter
            import numpy
            from numpy import float64
            """
        )

        for dtype in dtype_set:
            opt_choices = [
                itertools.combinations(opts.keys(), r=n)
                for n in range(1, len(opts) + 1)
            ]
            for opt_set in itertools.chain(*opt_choices):
                opt_kwarg = {k: opts[k] for k in opt_set if k}
                g = dict()
                exec(dim_test_imports, g)

                dim = tiledb.Dim(name="d1", dtype=dtype, **opt_kwarg)
                self.assertEqual(eval(repr(dim), g), dim)

        # test datetime
        g = dict()
        exec(dim_test_imports, g)
        dim = tiledb.Dim(
            name="d1",
            domain=(np.datetime64("2010-01-01"), np.datetime64("2020")),
            tile=2,
            dtype=np.datetime64("", "D"),
        )
        self.assertEqual(eval(repr(dim), g), dim)

    def test_arrayschema_repr(self, fx_sparse_cell_order):  # noqa: F811
        filters = tiledb.FilterList([tiledb.ZstdFilter(-1)])
        for sparse in [False, True]:
            cell_order = fx_sparse_cell_order if sparse else None
            domain = tiledb.Domain(
                tiledb.Dim(domain=(1, 8), tile=2), tiledb.Dim(domain=(1, 8), tile=2)
            )
            a1 = tiledb.Attr("val", dtype="f8", filters=filters)
            orig_schema = tiledb.ArraySchema(
                domain=domain, attrs=(a1,), sparse=sparse, cell_order=cell_order
            )

            schema_repr = repr(orig_schema)
            g = dict()
            setup = "from tiledb import *\n" "import numpy as np\n"

            exec(setup, g)
            new_schema = None
            try:
                new_schema = eval(schema_repr, g)
            except Exception:
                warn_str = (
                    """Exception during ReprTest schema eval"""
                    + """, schema string was:\n"""
                    + """'''"""
                    + """\n{}\n'''""".format(schema_repr)
                )
                warnings.warn(warn_str)
                raise

            self.assertEqual(new_schema, orig_schema)

    def test_arrayschema_repr_hilbert(self):
        domain = tiledb.Domain(tiledb.Dim(domain=(1, 8), tile=2))
        a = tiledb.Attr("a", dtype="f8")
        schema = tiledb.ArraySchema(
            domain=domain, attrs=(a,), cell_order="hilbert", sparse=True
        )

        assert schema.cell_order == "hilbert"
        assert schema.tile_order is None



from hypothesis import strategies as st
from hypothesis.strategies import composite

# Helpers for Hypothesis-Python based property tests
# (custom strategies, etc.)


@composite
def bounded_ntuple(draw, *, length=1, min_value=0, max_value=10):
    """hypothesis composite strategy that returns a `length` tuple of integers
    within the range (min_value, max_value)
    """

    return draw(st.tuples(*[st.integers(min_value, max_value) for _ in range(length)]))


@composite
def ranged_slices(draw, min_value=0, max_value=10):
    bdd = st.one_of(st.none(), st.integers(min_value=min_value, max_value=max_value))
    start = draw(bdd)
    stop = draw(bdd)
    step = draw(bdd)

    return slice(start, stop, step)



from typing import Any, Optional, Sequence, Union

import numpy as np

import tiledb.libtiledb as lt

from .ctx import Ctx, default_ctx
from .data_order import DataOrder
from .datatypes import DataType
from .filter import Filter, FilterList


class DimLabelSchema(lt.DimensionLabelSchema):
    def __init__(
        self,
        order: str = "increasing",
        label_dtype: np.dtype = np.uint64,
        dim_dtype: np.dtype = np.uint64,
        dim_tile: Any = None,
        label_filters: Union[FilterList, Sequence[Filter]] = None,
        ctx: Ctx = None,
    ):
        """Class defining a dimension label to create as part of an array.

        :param order: Order or sort of the label data ('increasing' or 'decreasing').
        :param label_dtype: Datatype of the label data.
        :param dim_dtype: Datatype of the target dimension.
        :param dim_tile: Tile extent for the dimension of the dimension label. If
            ``None``, it will use the same tile extent as the target dimension.
        :param label_filters: Filter list for the attribute storing the label data.
        :param ctx: A TileDB Context.
        """
        self._ctx = ctx or default_ctx()

        # Get DataType and DataOrder objects
        _label_order = DataOrder[order]
        _label_dtype = DataType.from_numpy(
            np.dtype(label_dtype)
            if label_dtype not in ("ascii", "blob")
            else label_dtype
        )
        np_dtype = _label_dtype.np_dtype
        if np.issubdtype(np_dtype, np.bytes_) or np.issubdtype(np_dtype, np.str_):
            _label_dtype = DataType(
                np_dtype, lt.DataType.STRING_ASCII, lt.TILEDB_VAR_NUM
            )

        _dim_dtype = DataType.from_numpy(dim_dtype)

        # Convert the tile extent (if set)
        if dim_tile is not None:
            if np.issubdtype(_dim_dtype.np_dtype, np.bytes_):
                raise TypeError(
                    "invalid tile extent, cannot set a tile a string dimension"
                )
            _dim_tile = _dim_dtype.cast_tile_extent(dim_tile)
        else:
            _dim_tile = None

        # Create the PyBind superclass
        if label_filters is None:
            super().__init__(
                _dim_dtype.tiledb_type,
                _dim_tile,
                _label_order.value,
                _label_dtype.tiledb_type,
            )
        else:
            _label_filters = (
                label_filters
                if isinstance(label_filters, FilterList)
                else FilterList(label_filters)
            )
            super().__init__(
                _dim_dtype.tiledb_type,
                _dim_tile,
                _label_order.value,
                _label_dtype.tiledb_type,
                _label_filters,
            )

    @property
    def dim_dtype(self) -> np.dtype:
        """Numpy dtype object representing the dimension type"""
        return DataType.from_tiledb(self._dim_dtype).np_dtype

    @property
    def dim_tile(self) -> Optional[np.generic]:
        """The tile extent of the dimension for the dimension label.

        :rtype: numpy scalar of np.timedelta64

        """
        tile_extent = self._dim_tile_extent
        if tile_extent is None:
            return None
        dim_dtype = DataType.from_tiledb(self._dim_dtype)
        return (
            None if tile_extent is None else dim_dtype.uncast_tile_extent(tile_extent)
        )

    @property
    def label_filters(self) -> Optional[FilterList]:
        """FilterList of the labels"""
        return (
            FilterList.from_pybind11(self._ctx, self._label_filters)
            if self._has_label_filters
            else None
        )

    @property
    def label_dtype(self) -> np.dtype:
        """Numpy dtype object representing the label type"""
        return DataType.from_tiledb(self._label_dtype).np_dtype

    @property
    def label_order(self) -> str:
        """Sort order of the labels on the dimension label"""
        return DataOrder(self._label_order).name
